from typing import List, Tuple, Union

import tensorflow as tf

from airontools.constructors.layers import layer_constructor


def block_constructor(
    units: List[int],
    input_shape: Tuple[int],
    name: str = "block",
    sequential: bool = False,
    bidirectional: bool = False,
    from_l: int = 1,
    hidden_activation: str = "prelu",
    output_activation: Union[str, tf.keras.layers.Layer] = "linear",
    kernel_regularizer_l1: float = 0.001,
    kernel_regularizer_l2: float = 0.001,
    bias_regularizer_l1: float = 0.001,
    bias_regularizer_l2: float = 0.001,
    dropout_rate: float = 0.0,
    bn: bool = False,
) -> tf.keras.models.Model:
    """It builds a custom block. For now only compatible with dense and sequential layers.

    Parameters:
        units (list): Number of units per hidden layer.
        input_shape (tuple): Input shape.
        name (str): Name of the block.
        sequential (bool): Whether to consider a sequential model or not.
        bidirectional (bool): Whether to consider bidirectional case or not (only active if sequential).
        from_l (int): The number indicator of the first hidden layer of the block, useful to make sure that layer
        names are not repeated.
        hidden_activation (str, Layer): Hidden activation function.
        output_activation (str, Layer): The activation function of the output of the block.
        dropout_rate (float): Probability of each intput being disconnected.
        kernel_regularizer_l1 (float): Kernel regularization using l1 penalization (Lasso).
        kernel_regularizer_l2 (float): Kernel regularization using l2 penalization (Ridge).
        bias_regularizer_l1 (float): Bias regularization using l1 penalization (Lasso).
        bias_regularizer_l2 (float): Bias regularization using l2 penalization (Ridge).
        bn (bool): If set, a batch normalization layer will be added right before the output activation function.

    Returns:
        model (Model): A keras model.
    """

    # Hidden layers
    i_l, o_l = (
        tf.keras.layers.Input(shape=input_shape, name="_".join([name, "input"])),
        None,
    )
    to_l = from_l + len(units)
    for l, o_dim in zip(range(from_l, to_l), units):
        if l <= from_l:
            o_l = i_l
        o_l = layer_constructor(
            x=o_l,
            units=o_dim,
            activation=hidden_activation
            if l == to_l - 1 is None
            else output_activation,
            name=name,
            name_ext=str(l),
            sequential=sequential,
            return_sequences=True if l < to_l - 1 and sequential else False,
            bidirectional=bidirectional,
            kernel_regularizer_l1=kernel_regularizer_l1,
            kernel_regularizer_l2=kernel_regularizer_l2,
            bias_regularizer_l1=bias_regularizer_l1,
            bias_regularizer_l2=bias_regularizer_l2,
            dropout_rate=dropout_rate,
            bn=bn,
        )

    # Model
    model = tf.keras.models.Model(inputs=i_l, outputs=o_l, name=name)

    return model



import warnings
from typing import Tuple, Union

import numpy as np
import tensorflow as tf

from airontools.constructors.utils import get_regularizer


def layer_constructor(
    x: Union[tf.Tensor, tf.keras.layers.Layer],
    units: int,
    name: str = "layer",
    name_ext: str = "",
    num_heads: int = 0,
    key_dim: int = 0,
    value_dim: int = 0,
    activation: Union[str, tf.keras.layers.Activation] = "prelu",
    use_bias: bool = True,
    sequential: bool = False,
    bidirectional: bool = False,
    return_sequences: bool = False,
    filters: int = 0,
    kernel_size: Union[int, Tuple[int]] = 0,
    padding: str = "valid",
    pooling: Union[str, tf.keras.layers.Layer] = None,
    pool_size: Union[int, Tuple[int]] = 1,
    conv_transpose: bool = False,
    strides: Union[int, Tuple[int]] = 1,
    sequential_axis: int = 1,
    kernel_regularizer_l1: float = 0.001,
    kernel_regularizer_l2: float = 0.001,
    bias_regularizer_l1: float = 0.001,
    bias_regularizer_l2: float = 0.001,
    dropout_rate: float = 0.0,
    bn: bool = False,
) -> tf.Tensor:
    """It builds a custom layer. For now only 2D convolutions
    are supported for input of rank 4.

        Parameters:
            x (tf.Tensor, tf.keras.layers.Layer): Input layer or tensor.
            units (int): Number of units for the dense layer. If a value is given, a dense layer will be added
            automatically if not sequential, else a sequential model. Useful to force an output dimensionality of the
            custom layer when using convolutional layers.
            name (str): Name of the custom layer.
            name_ext (str): Extension name for the custom layer that will be at the end of it.
            num_heads (int): Number of heads for the multi-head attention layer.
            key_dim (int): Key dimensionality for the multi-head attention layer, if None then the number of units is
            used instead.
            value_dim (int): Value dimensionality for the multi-head attention layer, if None then key_dim is used
            instead.
            activation (str, tf.keras.layers.Activation): The activation function of the output of the last hidden layer.
            use_bias (bool): Whether to sue bias or not.
            sequential (bool): Whether to consider a sequential custom layer or not. Sequential and self-attention
            (num_heads > 0) are not compatible.
            bidirectional (bool): Whether to consider bidirectional case or not (only active if sequential).
            names are not repeated.
            return_sequences (bool): Whether to return sequences or not (only active if sequential).
            filters (int): Number of filters to be used. If a value is given, a convolutional layer will be
            automatically added.
            kernel_size (int, Tuple[int]): Kernel size for the convolutional layer.
            conv_transpose (bool): Whether to use a transpose conv layer or not (only active if filters and
            kernel_size are set).
            strides (int, Tuple[int]): Strides for the conv layer (only active if filters and
            kernel_size are set).
            padding (str): Padding to be applied (only active if filters and
            kernel_size are set).
            pooling (str, tf.keras.layers.Layer): Pooling type.
            pool_size (int, Tuple[int]): Pooling size.
            sequential_axis (int): The axis that defines the sequence. For sequential models is normally 1. For
            self-attention (num_heads > 0) and image-like inputs, the sequential axis is the channel axis (3 for 2D
            images and 4 for 3D images).
            kernel_regularizer_l1 (float): Kernel regularization using l1 penalization (Lasso).
            kernel_regularizer_l2 (float): Kernel regularization using l2 penalization (Ridge).
            bias_regularizer_l1 (float): Bias regularization using l1 penalization (Lasso).
            bias_regularizer_l2 (float): Bias regularization using l2 penalization (Ridge).
            dropout_rate (float): Dropout rate.
            bn (bool): If set, a batch normalization layer will be added right before the output activation function.

        Returns:
            x (tf.keras.layers.Layer): A keras layer.
    """

    if num_heads > 0 and units is None and key_dim == 0:
        warnings.warn(
            "in order to use a multi-head attention layer either units or key_dim needs to be set",
        )

    # Dropout
    if dropout_rate != 0:
        x = dropout_layer_constructor(
            x,
            name=name,
            name_ext=name_ext,
            dropout_rate=dropout_rate,
        )

    # Convolution
    conv_condition = all(
        [conv_param != 0 for conv_param in [filters, kernel_size]],
    )
    if conv_condition:
        if isinstance(strides, int):
            _strides = tuple([strides] * _get_pooling_dim(x))
        else:
            _strides = strides
        conv_kwargs = dict(
            use_bias=use_bias,
            filters=filters,
            kernel_size=kernel_size,
            strides=_strides,
            padding=padding,
            kernel_regularizer=get_regularizer(
                kernel_regularizer_l1,
                kernel_regularizer_l2,
            ),
            bias_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
        )
        x = convolutional_layer_constructor(
            x,
            name=name,
            name_ext=name_ext,
            conv_transpose=conv_transpose,
            **conv_kwargs,
        )

    # Pooling
    if pooling is not None:
        if isinstance(pool_size, int):
            _pool_size = tuple([pool_size] * _get_pooling_dim(x))
        else:
            _pool_size = pool_size
        pooling_kwargs = dict(
            pool_size=_pool_size,
            strides=strides,
            padding=padding,
        )
        x = pooling_layer_constructor(
            x,
            name=name,
            name_ext=name_ext,
            conv_transpose=conv_transpose,
            pooling=pooling,
            **pooling_kwargs,
        )

    # Multi-Head Attention
    if num_heads > 0:
        _key_dim = key_dim if key_dim > 0 else units
        _value_dim = value_dim if value_dim > 0 else _key_dim
        multi_head_attention_kwargs = dict(
            num_heads=num_heads,
            key_dim=_key_dim,
            value_dim=_value_dim,
            use_bias=use_bias,
            kernel_regularizer=get_regularizer(
                kernel_regularizer_l1,
                kernel_regularizer_l2,
            ),
            bias_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
        )
        x = self_attention_layer_constructor(
            x,
            name=name,
            name_ext=name_ext,
            sequential_axis=sequential_axis,
            **multi_head_attention_kwargs,
        )

    # Sequential
    elif sequential:
        seq_kwargs = dict(
            units=units,
            use_bias=use_bias,
            kernel_regularizer=get_regularizer(
                kernel_regularizer_l1,
                kernel_regularizer_l2,
            ),
            bias_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
            return_sequences=return_sequences,
            activation="linear",
        )
        x = sequential_layer_constructor(
            x,
            name=name,
            name_ext=name_ext,
            bidirectional=bidirectional,
            sequential_axis=sequential_axis,
            **seq_kwargs,
        )

    # Dense
    if units > 0:
        dense_kwargs = dict(
            units=units,
            use_bias=use_bias,
            kernel_regularizer=get_regularizer(
                kernel_regularizer_l1,
                kernel_regularizer_l2,
            ),
            bias_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
        )
        x = dense_layer_constructor(
            x,
            name=name,
            name_ext=name_ext,
            **dense_kwargs,
        )

    # Batch Normalization
    if bn:
        bn_kwargs = dict(
            beta_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
            gamma_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
        )
        x = bn_layer_constructor(x, name=name, name_ext=name_ext, **bn_kwargs)

    # Activation
    activation_kwargs = dict(
        alpha_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
    )
    x = activation_layer_constructor(
        x,
        name=name,
        name_ext=name_ext,
        activation=activation,
        **activation_kwargs,
    )

    return x


def dropout_layer_constructor(
    x: Union[tf.Tensor, tf.keras.layers.Layer],
    dropout_rate: float,
    name: str = "dropout",
    name_ext: str = "",
) -> tf.keras.layers.Layer:
    input_shape = x.shape
    output_reshape = None
    if len(input_shape) > 2:
        output_reshape = input_shape
        x = tf.keras.layers.Flatten(
            name="_".join([name, "pre", "dropout", "flatten", name_ext])
        )(x)
    x = CustomDropout(name="_".join([name, "dropout", name_ext]), rate=dropout_rate)(x)
    if output_reshape is not None:
        x = tf.keras.layers.Reshape(
            name="_".join([name, "post", "dropout", "reshape", name_ext]),
            target_shape=output_reshape[1:],
        )(x)
    return x


def convolutional_layer_constructor(
    x: Union[tf.Tensor, tf.keras.layers.Layer],
    name: str = "convolution",
    name_ext: str = "",
    conv_transpose: bool = False,
    **kwargs,
) -> tf.keras.layers.Layer:
    assert len(x.shape) <= 3, "x layer shape should have 5 or less dimensions"
    conv_dim = len(x.shape) - 2
    conv_type = "transpose" if conv_transpose else ""
    conv_name = "conv" + str(conv_dim) + "d" + conv_type.capitalize()
    layer_name = "_".join([name, conv_name, name_ext])
    if conv_dim == 1:
        conv_layer = tf.keras.layers.Conv1D
    elif conv_dim == 2:
        conv_layer = tf.keras.layers.Conv2D
    else:
        conv_layer = tf.keras.layers.Conv3D
    x = conv_layer(
        name=layer_name,
        **kwargs,
    )(x)
    return x


def pooling_layer_constructor(
    x: Union[tf.Tensor, tf.keras.layers.Layer],
    name: str = "pooling",
    name_ext: str = "",
    pooling: Union[str, tf.keras.layers.Layer] = "max",
    **kwargs,
) -> tf.keras.layers.Layer:
    if isinstance(pooling, str):
        pooling_name = pooling.capitalize() + "Pooling" + str(_get_pooling_dim(x)) + "D"
        pooling_layer = globals()[pooling_name]
    else:
        pooling_name = [k for k, v in locals().iteritems() if v == pooling][0]
        pooling_layer = pooling
    x = pooling_layer(name="_".join([name, pooling_name.lower(), name_ext]))(x)
    return x


def _get_pooling_dim(x: tf.keras.layers.Layer) -> int:
    return len(x.shape) - 2


def self_attention_layer_constructor(
    x: Union[tf.Tensor, tf.keras.layers.Layer],
    name: str = "self_attention",
    name_ext: str = "",
    sequential_axis: int = 1,
    **kwargs,
) -> tf.keras.layers.Layer:
    x = sequential_permutation(
        x=x,
        name=name,
        name_ext=name_ext,
        sequential_axis=sequential_axis,
    )
    x = tf.keras.layers.MultiHeadAttention(
        name="_".join([name, "multi_head_attention", name_ext]),
        **kwargs,
    )(x, x)
    return x


def sequential_layer_constructor(
    x: Union[tf.Tensor, tf.keras.layers.Layer],
    name: str = "sequential",
    name_ext: str = "",
    bidirectional: bool = False,
    sequential_axis: int = 1,
    **kwargs,
) -> tf.keras.layers.Layer:
    x = sequential_permutation(
        x=x,
        name=name,
        name_ext=name_ext,
        sequential_axis=sequential_axis,
    )
    if bidirectional:
        x = tf.keras.layers.Bidirectional(
            tf.keras.layers.GRU(name="_".join([name, "gru", name_ext]), **kwargs)
        )(x)
    else:
        x = tf.keras.layers.GRU(name="_".join([name, "gru", name_ext]), **kwargs)(x)
    return x


def dense_layer_constructor(
    x: Union[tf.Tensor, tf.keras.layers.Layer],
    name: str = "dense",
    name_ext: str = "",
    **kwargs,
) -> tf.keras.layers.Layer:
    if not len(x.shape[1:]) == 1:
        x = tf.keras.layers.Flatten(
            name="_".join([name, "pre", "dense", "flatten", name_ext])
        )(x)
    x = tf.keras.layers.Dense(name="_".join([name, "dense", name_ext]), **kwargs)(x)
    return x


def bn_layer_constructor(
    x: Union[tf.Tensor, tf.keras.layers.Layer], name: str, name_ext: str, **kwargs
) -> tf.keras.layers.Layer:
    input_shape = x.shape
    output_reshape = None
    if len(input_shape) > 2:
        output_reshape = input_shape
        x = tf.keras.layers.Flatten(
            name="_".join([name, "pre", "bn", "flatten", name_ext])
        )(x)
    x = tf.keras.layers.BatchNormalization(
        name="_".join([name, "batch_normalization", name_ext]),
        **kwargs,
    )(x)
    if output_reshape is not None:
        x = tf.keras.layers.Reshape(
            name="_".join([name, "post", "bn", "reshape", name_ext]),
            target_shape=output_reshape[1:],
        )(x)
    return x


def activation_layer_constructor(
    x: Union[tf.Tensor, tf.keras.layers.Layer],
    name: str = "activation",
    name_ext: str = "",
    activation: str = "prelu",
    **reg_kwargs,
) -> tf.keras.layers.Layer:
    input_shape = x.shape
    output_reshape = None
    if len(input_shape) > 2:
        output_reshape = input_shape
        x = tf.keras.layers.Flatten(
            name="_".join([name, "pre", "activation", "flatten", name_ext])
        )(x)
    activation_name = "_".join(
        [name, activation if isinstance(activation, str) else "activation", name_ext],
    )
    if activation == "leakyrelu":
        x = tf.keras.layers.LeakyReLU(name=activation_name)(x)
    elif activation == "prelu":
        x = tf.keras.layers.PReLU(name=activation_name, **reg_kwargs)(x)
    elif activation == "softmax":
        x = tf.keras.layers.Softmax(name=activation_name, dtype="float32")(x)
    else:
        x = tf.keras.layers.Activation(name=activation_name, activation=activation)(x)
    if output_reshape is not None:
        x = tf.keras.layers.Reshape(
            name="_".join([name, "post", "activation", "reshape", name_ext]),
            target_shape=output_reshape[1:],
        )(x)
    return x


def sequential_permutation(
    x: Union[tf.Tensor, tf.keras.layers.Layer],
    name: str = "permutation",
    name_ext: str = "",
    sequential_axis: int = 1,
) -> tf.keras.layers.Layer:
    input_shape = x.shape
    sequential_axis_ = list(range(len(input_shape)))[sequential_axis]
    if sequential_axis_ != 1:
        permutation = tuple(
            [sequential_axis_]
            + [i for i in range(1, len(input_shape[1:])) if i != sequential_axis_],
        )
        x = tf.keras.layers.Permute(
            name="_".join([name, "permutation", name_ext]), dims=permutation
        )(x)
    else:
        permutation = list(range(1, len(input_shape)))
    if len(input_shape) > 2:
        x = tf.keras.layers.Reshape(
            name="_".join([name, "reshape", name_ext]),
            target_shape=(
                input_shape[permutation[0]],
                np.prod([input_shape[i] for i in permutation[1:]]),
            ),
        )(x)
    return x


def identity(x: tf.keras.layers.Layer) -> tf.keras.layers.Layer:
    """Identity layer."""
    return x


class CustomDropout(tf.keras.layers.Dropout):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.rate = tf.Variable(
            self.rate,
            trainable=False,
            name="_".join([self.name, "rate"]),
            dtype=self.dtype,
        )



from typing import List

from tensorflow.python.client import device_lib


def get_available_gpus() -> List[str]:
    """Gets available devices.
    Returns:
        A list of device names.
    """
    local_device_protos = device_lib.list_local_devices()
    return [x.name for x in local_device_protos if x.device_type == "GPU"]



import os
import random
import tempfile
from random import seed

import numpy as np
import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from sklearn.model_selection import KFold
from tensorflow import DType


def train_val_split(
    input_data,
    output_data=None,
    meta_data=None,
    n_parallel_models=1,
    do_kfolds=False,
    val_ratio=0.2,
    shuffle=True,
    seed_val=None,
    return_tfrecord=False,
    tfrecord_name=None,
):
    """Train validation split.

    Parameters:
        input_data (list[array, tf.data.Dataset], array, tf.data.Dataset): Input data.
        output_data (list[array, tf.data.Dataset], array, tf.data.Dataset): Output data.
        meta_data (list[array, tf.data.Dataset], array, tf.data.Dataset): Meta data.
        n_parallel_models (int): Number of parallel models.
        do_kfolds (bool): Whether to do kfolds for cross-validation or not.
        val_ratio (float): Ratio for validation.
        shuffle (bool): Whether to shuffle or not.
        seed_val (int): Seed value.
        return_tfrecord (bool): Whether to return tfrecord or not.
        tfrecord_name (str): Name of the tfrecord.

    Returns:
        4 list[array, tf.data.Dataset].
    """
    distributions = ["train", "val"]
    data = dict(x=_to_list_array(input_data))
    n_samples = data["x"][0].shape[0]
    split_data = dict(x={distribution: [] for distribution in distributions})
    if output_data is not None:
        data.update(y=_to_list_array(output_data))
        split_data.update(y={distribution: [] for distribution in distributions})
    if meta_data is not None:
        data.update(meta=_to_list_array(meta_data))
        split_data.update(meta={distribution: [] for distribution in distributions})
    if do_kfolds and n_parallel_models > 1:
        kf = KFold(n_splits=n_parallel_models, shuffle=True, random_state=seed_val)
        n_train = min(data_[0].shape[0] for data_ in kf.split(range(n_samples)))
        n_val = min(data_[1].shape[0] for data_ in kf.split(range(n_samples)))
        inds = [
            dict(train=train_inds[:n_train, ...], val=val_inds[:n_val, ...])
            for train_inds, val_inds in kf.split(range(n_samples))
        ]
    else:
        inds = list(np.arange(n_samples))
        if shuffle:
            random.shuffle(inds, random=seed(seed_val))
        line = int(len(inds) * (1 - val_ratio))
        inds = [dict(train=inds[:line], val=inds[line:])] * n_parallel_models
    for inds_ in inds:
        for distribution in distributions:
            x_ = []
            for sub_x in data["x"]:
                x_ += [sub_x[inds_[distribution], ...]]
            split_data["x"][distribution] += [x_]
            if output_data is not None:
                y_ = []
                for sub_y in data["y"]:
                    y_ += [sub_y[inds_[distribution], ...]]
                split_data["y"][distribution] += [y_]
            if meta_data is not None:
                meta_ = []
                for sub_meta in data["meta"]:
                    meta_ += [sub_meta[inds_[distribution], ...]]
                split_data["meta"][distribution] += [meta_]
    if return_tfrecord:
        if tfrecord_name is None:
            tfrecord_name = os.path.join(tempfile.gettempdir(), "tfrecord")
        for name in split_data.keys():
            for distribution in distributions:
                for i in range(n_parallel_models):
                    for j in range(len(split_data[name][distribution][i])):
                        tfrecord_name_ = "_".join(
                            [tfrecord_name, name, distribution, "fold", str(i), str(j)],
                        )
                        _data = split_data[name][distribution][i][j]
                        sample_shape = tuple(_data.shape[1:])
                        write_tfrecord(data=_data, name=tfrecord_name_ + ".tfrecords")
                        split_data[name][distribution][i][j] = read_tfrecord(
                            name=tfrecord_name_ + ".tfrecords",
                            sample_shape=sample_shape,
                        )
    returns = []
    for name in split_data.keys():
        for distribution in distributions:
            return_ = split_data[name][distribution]
            if len(data[name]) == 1:
                return_ = [_return[0] for _return in return_]
            if n_parallel_models == 1:
                return_ = return_[0]
            returns += [return_]
    returns += [inds]
    return returns


def to_time_series(dataset, targets, look_back=1):
    union_dataset = np.concatenate((dataset, targets), axis=-1)
    x, y = [], []
    for i in range(len(union_dataset) - look_back - 1):
        x.append(union_dataset[i : (i + look_back), ...])
        y.append(targets[i + look_back, ...])
    return np.array(x), np.array(y)


def write_tfrecord(data: np.array, name: str):
    with tf.io.TFRecordWriter(name) as writer:
        for i in range(len(data)):
            sample = data[i].reshape((np.prod(data[i].shape),)).astype(np.float32)
            example = _example(sample)
            writer.write(example.SerializeToString())


def read_tfrecord(name: str, sample_shape: tuple, dtype=tf.float32):
    dataset = tf.data.TFRecordDataset(name)
    dataset = dataset.map(parse_function(sample_shape=sample_shape, dtype=dtype))
    return dataset


def parse_function(sample_shape, dtype: DType):
    def parse_function_(record):
        feature_description = {"data": tf.io.FixedLenFeature([], tf.string)}
        data = tf.io.parse_single_example(record, feature_description)
        data = tf.io.decode_raw(data["data"], out_type=dtype)
        data = tf.reshape(data, sample_shape)
        return data

    return parse_function_


def _example(data: np.array):
    feature = {"data": _bytes_feature(data.tobytes())}
    features = tf.train.Features(feature=feature)
    example = tf.train.Example(features=features)
    return example


def _bytes_feature(value):
    """Returns a bytes_list from a string / byte."""
    if isinstance(value, type(tf.constant(0))):
        # BytesList won't unpack a string from an EagerTensor.
        value = value.numpy()
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint."""
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def _to_list_array(data):
    data_list = []
    if not isinstance(data, list):
        data = [data]
    for i in range(len(data)):
        if isinstance(data[i], tf.data.Dataset):
            data[i] = tfds.as_numpy(data[i])
            for _, data_ in data[i].items():
                data_list += [data_]
        else:
            data_list += [data[i]]
    return data_list



import warnings

import numpy as np
import tensorflow as tf
import tensorflow.python.keras.backend as k_bcknd


def get_latent_model(model: tf.keras.models.Model, layer_name: str):
    try:
        return tf.keras.models.Model(
            inputs=model.inputs,
            outputs=model.get_layer(layer_name).output,
            name=layer_name + "_model",
        )
    except ValueError:
        outputs = model.inputs
        layer_found = False
        for layer in model.layers:
            outputs = layer(outputs)
            if layer.name == layer_name:
                layer_found = True
                break
        if layer_found:
            return tf.keras.models.Model(
                inputs=model.inputs,
                outputs=outputs,
                name=layer_name + "_model",
            )
        else:
            warnings.warn("could not find the layer")


def set_precision(precision: str) -> None:
    if "float16" in precision:
        if precision == "mixed_float16":
            policy = tf.keras.mixed_precision.Policy("mixed_float16")
            tf.keras.mixed_precision.set_global_policy(policy)
        else:
            tf.keras.backend.set_floatx("float16")


def to_time_series(tensor: tf.Tensor) -> tf.Tensor:
    return k_bcknd.expand_dims(tensor, axis=2)


def get_layer_units(
    input_dim: int,
    output_dim: int,
    n_layers: int,
    min_hidden_units=2,
) -> list:
    units = [
        max(int(units), min_hidden_units)
        for units in np.linspace(input_dim, output_dim, n_layers + 1)
    ]
    units[0], units[-1] = input_dim, output_dim
    return units


def rm_redundant(values: list, value: int) -> list:
    taken = False
    values_ = []
    for n in values:
        if n != value:
            values_ += [n]
        elif not taken:
            values_ += [n]
            taken = True
    return values_


def get_regularizer(
    l1_value: float = None, l2_value: float = None
) -> tf.keras.regularizers.Regularizer:
    if l1_value and l2_value:
        return tf.keras.regularizers.l1_l2(l1=l1_value, l2=l2_value)
    elif l1_value:
        return tf.keras.regularizers.l1(l1_value)
    elif l2_value:
        return tf.keras.regularizers.l2(l2_value)



import json
from typing import Dict, Tuple

import numpy as np
import tensorflow as tf
from numpy.typing import NDArray

from airontools.constructors.layers import layer_constructor
from airontools.constructors.models.model import Model
from airontools.on_the_fly import HyperDesignDropoutRate


class AE(Model, tf.keras.models.Model):
    """AutoEncoder model."""

    def __init__(
        self,
        input_shape: Tuple[int],
        model_name: str = "AE",
        output_activation: str = "softmax",
        latent_dim: int = 3,
        **kwargs,
    ):
        Model.__init__(self)
        tf.keras.models.Model.__init__(self)

        # Loss tracker
        self.loss_tracker = tf.keras.metrics.Mean(name="loss")

        # Encoder
        encoder_inputs = tf.keras.layers.Input(shape=input_shape)
        self.encoder = layer_constructor(
            encoder_inputs,
            units=latent_dim,
            name=f"{model_name}_encoder",
            **kwargs,
        )
        self.encoder = tf.keras.models.Model(
            inputs=encoder_inputs,
            outputs=self.encoder,
            name=f"{model_name}_encoder",
        )

        # Z
        z_inputs = tf.keras.layers.Input(shape=(latent_dim,))
        self.z = layer_constructor(
            z_inputs,
            units=latent_dim,
            name=f"{model_name}_z",
            **kwargs,
        )
        self.z = tf.keras.models.Model(
            inputs=z_inputs,
            outputs=self.z,
            name=f"{model_name}_z",
        )

        # Decoder
        decoder_inputs = tf.keras.layers.Input(shape=(latent_dim,))
        self.decoder = layer_constructor(
            decoder_inputs,
            units=self.encoder.input_shape[-1],
            name=f"{model_name}_decoder",
            activation=output_activation,
            **kwargs,
        )
        self.decoder = tf.keras.models.Model(
            inputs=decoder_inputs,
            outputs=self.decoder,
            name=f"{model_name}_decoder",
        )

        # AE
        self._model = tf.keras.models.Model(
            inputs=encoder_inputs,
            outputs=self.decoder(self.z(self.encoder(encoder_inputs))),
            name=model_name,
        )

        # Hyper design on the fly
        self.hyper_design_dropout_rate = HyperDesignDropoutRate(model=self._model)

    def compile(self, *args, **kwargs) -> None:
        """Compile model."""
        tf.keras.models.Model.compile(self, *args, **kwargs)

    def fit(self, *args, **kwargs) -> None:
        """Compile model."""
        tf.keras.models.Model.fit(self, *args, **kwargs)

    def evaluate(self, x: NDArray[float], **kwargs) -> Dict[str, tf.Tensor]:
        """Evaluate model."""
        reconstructed = self._model(x)
        loss = self._loss_evaluation(reconstructed, x)
        return {"loss": loss}

    def predict(self, *args, **kwargs) -> NDArray[float]:
        """Predict."""
        return tf.keras.models.Model.predict(self, *args, **kwargs)

    def save_weights(self, path: str) -> None:
        """Save model weights."""
        with open(path + "_weights", "w") as f:
            json.dump([w.tolist() for w in self._model.get_weights()], f)

    def load_weights(self, path: str) -> None:
        """Load model weights."""
        with open(path + "_weights") as f:
            encoder_weights = [np.array(w) for w in json.load(f)]
        self._model.set_weights(encoder_weights)

    def call(self, inputs, **kwargs) -> tf.Tensor:
        """Call model."""
        reconstructed = self._model(inputs)
        self.add_loss(self._loss_evaluation(reconstructed, inputs))
        return reconstructed

    def summary(self, **kwargs) -> None:
        """Model summary."""
        self._model.summary(**kwargs)

    def _loss_evaluation(
        self, reconstructed: tf.Tensor, inputs: tf.Tensor
    ) -> tf.Tensor:
        rec_loss = tf.reduce_mean((inputs - reconstructed) ** 2)
        return rec_loss



import os
import shutil


def path_management(path: str, modes=None) -> None:
    """Path preparation.

    Parameters:
        path (str): Path to manage.
        modes (list): List of mode per path, remove as 'rm' and make as 'make'.
    """
    available_modes = ["rm", "make"]
    if not modes:
        modes = ["make"]
    for mode in modes:
        assert mode in available_modes
        if os.path.isdir(path) and mode == "rm":
            shutil.rmtree(path)
        elif mode == "make":
            os.makedirs(path, exist_ok=True)



import os
from typing import List, Union

import tensorflow as tf
from numpy.typing import NDArray
from tensorboard.plugins import projector

from airontools.tools import path_management


def save_representations(
    representations: NDArray[float],
    path: str,
    representations_name: str = "representations",
    metadata: Union[
        List[NDArray[Union[float, str]]], NDArray[Union[float, str]]
    ] = None,
) -> None:
    """Save representations (or embeddings).

    Parameters:
        representations (NDArray[float]): Representations to be saved.
        path (str): Path to save the representations.
        representations_name (str): Embeddings names.
        metadata (Union[List[NDArray[Union[float, str]]], NDArray[Union[float, str]]]): Metadata (a list of arrays or an array).
    """

    # Path management
    path_management(path)

    # Save metadata
    metadata_file_name = os.path.join(path, "metadata.tsv")
    if metadata is not None:
        metadata_list = metadata if isinstance(metadata, list) else [metadata]
        with open(metadata_file_name, "w") as f:
            for i in range(len(metadata_list[0])):
                meta_line = []
                for metadata_ in metadata_list:
                    meta_line += [str(metadata_[i])]
                f.write("\t".join(meta_line) + "\n")

    # Save representations
    representations_var = tf.Variable(representations, name=representations_name)
    checkpoint = tf.train.Checkpoint(embedding=representations_var)
    checkpoint.save(os.path.join(path, representations_name + ".ckpt"))

    # Set up config and representations
    config = projector.ProjectorConfig()
    embedding = config.embeddings.add()
    embedding.tensor_name = "embedding/.ATTRIBUTES/VARIABLE_VALUE"
    embedding.metadata_path = metadata_file_name
    projector.visualize_embeddings(path, config)



import warnings

import tensorflow as tf


class HyperDesignDropoutRate:
    """Hyper-parameter designer for the dropout rate on the fly."""

    def __init__(
        self, model: tf.keras.models.Model, down: float = -0.01, up: float = 0.01
    ):
        self.rates = []
        for layer in model.layers:
            if isinstance(layer, tf.keras.models.Model):
                for sub_layer in layer.layers:
                    self.__append_rate(sub_layer)
            elif isinstance(layer, tf.keras.layers.Layer):
                self.__append_rate(layer)
        self.actions_space = {}
        for action_name, action_value in zip(["down", "stay", "up"], [down, 0.0, up]):
            self.actions_space.update(
                {
                    action_name: tf.constant(
                        action_value,
                        dtype=model.dtype,
                        name="_".join(["rate", action_name]),
                    ),
                },
            )

    def __append_rate(self, layer: tf.keras.layers.Layer) -> None:
        if hasattr(layer, "rate"):
            if isinstance(layer.rate, tf.Variable):
                self.rates += [layer.rate]
            else:
                warnings.warn(
                    "layer {} does not contain a rate as a tf.Variable".format(
                        layer.name,
                    ),
                )

    def set_action(self, action: str) -> None:
        """Set action for the dropout rate. Whether to go down, stay or up."""
        assert action in ["down", "stay", "up"]
        for rate in self.rates:
            new_rate = rate + self.actions_space[action]
            new_rate_ = tf.keras.backend.get_value(new_rate)
            if 0 <= new_rate_ < 1:
                tf.keras.backend.set_value(rate, new_rate)
